/**
 * cxline.cpp:
 *
 * Copyright (c) Hongfeng Yu <hfstudio@gmail.com>
 *
 * All rights reserved.  May not be used, modified, or copied
 * without permission.
 *
 */
#include <cassert>
#include "cxline.h"

using namespace std;

cxLine::cxLine()
{
    Clear();
}

cxLine::~cxLine()
{
    m_vRawLine.clear();
}

void cxLine::Clear() 
{
    m_vRawLine.clear();
    m_vParaLine.clear();
    m_vLenPercent.clear();
    m_fTotalLength = 0;
    m_bUpdateLenPercent = true;
    m_bUpdateSamples = true;
    m_bUpdatePort = true;
}

void cxLine::PushBack(cxLinePoint p) 
{
    m_vRawLine.push_back(p);
    
    m_bUpdateLenPercent = true;
    m_bUpdateSamples = true;
    
    int size = m_vRawLine.size(); 
    if ( size > 1) {
        cx3DVector curt = m_vRawLine[size - 1].Pos();
        cx3DVector prev = m_vRawLine[size - 2].Pos();
        float temp = curt.distance(prev);
        m_fTotalLength += temp;
    }
}

void cxLine::PushFront(cxLinePoint p) 
{
    m_vRawLine.insert(m_vRawLine.begin(), p);
    
    m_bUpdateLenPercent = true;
    m_bUpdateSamples = true;
    
    int size = m_vRawLine.size();
    if ( size > 1) {
        cx3DVector curt = m_vRawLine[0].Pos();
        cx3DVector prev = m_vRawLine[1].Pos();
        float temp = curt.distance(prev);
        m_fTotalLength += temp;
    }
}

void cxLine::UpdateLenPercent()
{
    if (m_vRawLine.size() < 2) {
        return;
    }
    
    if (m_fTotalLength == 0) {
        return;
    }
        
    // calculate the percentage
    m_vLenPercent.push_back(0);
    float accum_len = 0;
    int size = m_vRawLine.size();
    for (int i = 1; i < size; i++) {
        cx3DVector curt = m_vRawLine[i].Pos();
        cx3DVector prev = m_vRawLine[i - 1].Pos();
        float temp = curt.distance(prev);
        accum_len += temp;
        m_vLenPercent.push_back(accum_len / m_fTotalLength);
    }

    float max = *(m_vLenPercent.rbegin());
    
    if ( fabs(max - 1.0f) > 0.0001) {
        cerr << "the maximum percentage is " << max << endl;
        assert(0);
    }
    
    m_bUpdateLenPercent = false;
}


cxLinePoint cxLine::LenPercentInterPoint(float percent)
{
    assert(percent >= 0.0 && percent <= 1.0);

    if (m_bUpdateLenPercent) {
        UpdateLenPercent();
    }

    // the maximum percent could be smaller than 1 
    if (percent > *(m_vLenPercent.rbegin())) {
        percent = *(m_vLenPercent.rbegin());
    }

    int left, right, mid;
    left = 0;
    right = m_vLenPercent.size() - 1;
    
    while((right - left) > 1){
        mid = (left + right) / 2;
        if (m_vLenPercent[mid] >= percent) {
            right = mid;
        } else if (m_vLenPercent[mid] < percent) {
            left = mid;
        }
    }

    assert(percent >= m_vLenPercent[left]);
    assert(percent <= m_vLenPercent[right]);
    assert(right == left + 1);
           
    float f = (percent - m_vLenPercent[left]) /
              (m_vLenPercent[right] - m_vLenPercent[left]);
    
    cxLinePoint point;

    point = m_vRawLine[left] * (1.0f - f) + m_vRawLine[right] * f;

    return point;
}

void cxLine::Parameterize()
{
    m_vParaLine.clear();
    
    if (m_vRawLine.size() < 2) {
        return;
    }
    
    if (m_bUpdateLenPercent) {
        UpdateLenPercent();
    }
    
    float accum_len = 0;
    float percent = 0;
    while (accum_len <= m_fTotalLength)
    {
        percent = accum_len / m_fTotalLength;
        cxLinePoint pos = LenPercentInterPoint(percent);
        m_vParaLine.push_back(pos);
        accum_len += PARA_INTERVAL;
    }
    
    if (percent < 1.0) {
        percent = 1.0;
        cxLinePoint pos = LenPercentInterPoint(percent);
        m_vParaLine.push_back(pos);
    }
}

void cxLine::UpdateSamples()
{
    m_vSamples.clear();
    
    int parasize = m_vParaLine.size();
    
    if (parasize <= SAMPLE_NUMBER) 
    {
        for (int i = 0; i < parasize; i++)
        {
            m_vSamples.push_back(m_vParaLine[i].Pos());
        }
    } else {
        for (int i = 0; i < SAMPLE_NUMBER; i++) {
            float pos = (parasize - 1) * i * 1.0f / SAMPLE_NUMBER;
            float f = (pos - (int)pos);
            int left = (int)(pos);
            int right = (int)(pos) + 1;
            if (right >= parasize - 1) {
                right = parasize - 1;
            }
            
            cxLinePoint point;
            point = m_vParaLine[left] * (1.0f - f) + m_vParaLine[right] * f;
            m_vSamples.push_back(point);
        }
   }
    
    m_bUpdateSamples = false;
}

void cxLine::UpdatePort()
{
    int parasize = m_vParaLine.size();
    float minx, maxx, miny, maxy, minz, maxz;
    minx = miny = minz = 1.0E10;
    maxx = maxy = maxz = -1.0E10;
    
    for (int i = 0; i < parasize; i++) {
        cx3DVector vect = m_vParaLine[i].Pos();
        if (minx > vect.x()) minx = vect.x();
        if (maxx < vect.x()) maxx = vect.x();
        
        if (miny > vect.y()) miny = vect.y();
        if (maxy < vect.y()) maxy = vect.y();
        
        if (minz > vect.z()) minz = vect.z();
        if (maxz < vect.z()) maxz = vect.z();        
    }
    
    m_fPort[0] = minx;
    m_fPort[1] = maxx;
    
    m_fPort[2] = miny;
    m_fPort[3] = maxy;
    
    m_fPort[4] = minz;
    m_fPort[5] = maxz;
    
    m_bUpdatePort = false;
}

void cxLine::Write(ofstream &outf)
{
    int line_size = m_vParaLine.size();
    outf.write(reinterpret_cast<const char*>(&line_size), sizeof(int));
    
    for (int i = 0; i < line_size; i++) {
        cxLinePoint point = m_vParaLine[i];
        outf.write(reinterpret_cast<const char*>(&point),
                   sizeof(cxLinePoint));
    }
    
    outf.write(reinterpret_cast<const char*>(&m_SeedPoint), 
               sizeof(cx3DVector));
}

void cxLine::Read(ifstream &inf)
{
    int line_size;
    inf.read(reinterpret_cast<char *>(&line_size), sizeof(int));

    for (int j = 0; j < line_size; j++) {
        cxLinePoint point;
        inf.read(reinterpret_cast<char *>(&point), sizeof(cxLinePoint));
        m_vParaLine.push_back(point);
    }
    
    inf.read(reinterpret_cast<char*>(&m_SeedPoint), 
             sizeof(cx3DVector));
}

void cxLine::IntersectHorizon(float y, float minx, float maxx, 
                              vector<cx3DVector> &ins_pos)
{
    int size = m_vParaLine.size() - 1;
    
    float x1, y1, x2, y2, x, f;
    
    for (int i = 0; i < size; i++) {
        
        cxLinePoint &curt = m_vParaLine[i];
        cxLinePoint &next = m_vParaLine[i+1];

        x1 = curt.Pos().x();
        y1 = curt.Pos().y();

        x2 = next.Pos().x();
        y2 = next.Pos().y();
        
        if (((y1 >= y && y2 <= y) || (y1 <= y && y2 >= y)) &&
            !(x1 < minx && x2 < minx) &&
            !(x1 > maxx && x2 > maxx))
        {
            if (y1 == y2 && y1 == y) 
            {
                if ((minx >= x1 && minx <= x2) ||
                    (minx >= x2 && minx <= x1))
                {
                    //exclude next point
                    if (minx != x2 && y != y2) {
                        ins_pos.push_back(cx3DVector(minx, y, 0));
                    }
                }
                
                if ((maxx >= x1 && maxx <= x2) ||
                    (maxx >= x2 && maxx <= x1))
                {
                    //exclude next point
                    if (maxx != x2 && y != y2) {
                        ins_pos.push_back(cx3DVector(maxx, y, 0));
                    }
                }
            } else {
                
                f = (y - y1) / (y2 - y1);
                x = (x2 - x1) * f + x1;
                
                if (x >= minx && x <= maxx) {
                    //exclude next point
                    if (x != x2 && y != y2) {
                        ins_pos.push_back(cx3DVector(x, y, 0));
                    }
                }
            }
        }
    }
}
    

void cxLine::IntersectVertical(float x, float miny, float maxy, 
                               vector<cx3DVector> &ins_pos)
{

    int size = m_vParaLine.size() - 1;
    
    float x1, y1, x2, y2, y, f;
    
    for (int i = 0; i < size; i++) {
        
        cxLinePoint &curt = m_vParaLine[i];
        cxLinePoint &next = m_vParaLine[i+1];
        
        x1 = curt.Pos().x();
        y1 = curt.Pos().y();
        
        x2 = next.Pos().x();
        y2 = next.Pos().y();
        
        if (((x1 >= x && x2 <= x) || (x1 <= x && x2 >= x)) &&
            !(y1 < miny && y2 < miny) &&
            !(y1 > maxy && y2 > maxy))
        {
            if (x1 == x2 && x1 == x) 
            {
                if ((miny >= y1 && miny <= y2) ||
                    (miny >= y2 && miny <= y1))
                {
                    //exclude next point
                    if (x != x2 && miny != y2) {
                        ins_pos.push_back(cx3DVector(x, miny, 0));
                    }
                }
                
                if ((maxy >= y1 && maxy <= y2) ||
                    (maxy >= y2 && maxy <= y1))
                {
                    //exclude next point
                    if (x != x2 && maxy != y2) {
                        ins_pos.push_back(cx3DVector(x, maxy, 0));
                    }
                }
            } else {
                
                f = (x - x1) / (x2 - x1);
                y = (y2 - y1) * f + y1;
                
                if (y >= miny && y <= maxy) {
                    //exclude next point
                    if (x != x2 && y != y2) {
                        ins_pos.push_back(cx3DVector(x, y, 0));
                    }
                }
            }
        }
    }
}




